#!/usr/bin/perl

# [en] if this script fails to run in mod_perl with errors like
# [de] wenn dieses Skript unter mod_perl mit diesen Fehlern abbricht
# * Premature end of script headers: cdn.cgi
# * No such file or directory: exec of '/var/www/cgi-bin/cdn.cgi' failed
# [en] then something is wrong with CR/LF, so run once:
# [de] dann stimmt etwas nicht mit den Zeilenumbrüchen. Bitte ausführen:
# $ dos2unix /var/www/cgi-bin/cdn.cgi

=head1 NAME

cdn.cgi - Content Delivery Network Web-UI

=head1 DESCRIPTION

[en] C<cdn.cgi> is a minimal web page to quickly gather major information
and software states used by XtreemFS, PowerDNS and lsyncd.
[de] C<cdn.cgi> ist eine kleine Webseite zum Darstellen von verschiedenen
Information über XtreemFS, PowerDNS und lsyncd.

=head1 INSTALLATION

[en] Move this file to the C<cgi-bin> directory of your web server and
access the URL from any web browser. It requires Perl environment
within the web server and requires CGI, Geo::IP and Sys::Hostname
modules.
[de] Kopiere diese Datei ins C<cgi-bin> Verzeichnis des Webserver und
rufe die URL in einem Webbrowser auf. Das Skript benötigt eine Perl-
umgebung im Webserver mit den Modulen CGI, Geo::IP und Sys::Hostname

=head1 REQUIREMENTS

[en] This script is optimized to work with the pathnames used in CentOS
6.8 but other distributions should be fine if you adjust the names
in the variables below.
[de] Dieses Skript ist angepasst an die Pfadnamen von CentOS 6.8, aber
andere Distributionen sollten auch funktioneren, solange die Pfade ent-
sprechend angepasst werden.

=head1 SUDO

[en] C<cdn.cgi> tries to execute some commands with require root
privileges. It uses sudo, so please add the following lines to your sudo
configuration, where 'apache' is the owner of the apache process.
[de] C<cdn.cgi> muss verschiedene Kommandos mit root-Rechten ausführen.
Dazu sind die folgenden Zeilen in der sudo-Konfiguration notwendig, wobei
'apache' der Besitzer des apache-Prozesses ist.

Defaults    !requiretty
apache  ALL=(ALL)  NOPASSWD:/usr/bin/xtfsutil *
apache  ALL=(ALL)  NOPASSWD:/usr/bin/lsfs.xtreemfs *
apache  ALL=(ALL)  NOPASSWD:/sbin/mii-tool *
apache  ALL=(ALL)  NOPASSWD:/usr/bin/pdns_control *

=cut

# [en]	path to local files
# [de]	Pfade zu den lokalen Dateien
my $path_to_xtfsutil	= '/usr/bin/xtfsutil';
my $path_to_lsfs	= '/usr/bin/lsfs.xtreemfs';
my $path_to_miitool	= '/sbin/mii-tool';
my $path_to_pdns_control= '/usr/bin/pdns_control';
my $lsyncd_status_file	= '/tmp/lsyncd.status';
my $conf_softflowd	= '/etc/sysconfig/softflowd';
my $url_country_flag	= 'http://lipis.github.io/flag-icon-css/flags/4x3';

# [en]	how many lines to show from a single log file
# [de]	Wieviele Zeilen sollen von einer Logdatei maximal angezeigt werden?
my $logfile_max_lines	= 1000;


# ---- Used modules (main program)
use CGI;
use Sys::Hostname qw//;
use Geo::IP;

# ---- declarations ----
sub quitProgram( $ );
sub timef( $ );
sub get_file_content( $@ );
sub print_cli_output( $$$ );


# [en]	colors for web page (based on www.colorbrewer2.org)
# [de]	Farben für die Webseite (basiert auf www.colorbrewer2.org)
my $color_one	= '#4d4d4d';	# rdgy-11-div-10 (gray)
my $color_two	= '#9970ab';	# prgn-11-div-3  (purple)
my $color_light	= '#f7f7f7';	# prgn-11-div-6
my $color_dark	= '#4d4d4d';	# rdgy-11-div-10


# [en]	build CGI/HTML header
# [de]	CGI/HTML Kopfzeilen erstellen
my $q = CGI->new;
my $section = defined $q ? $q->param('section') : '';
print $q->header( 'text/html' );

# [en]	poor-mans authentication
# [de]	eine sehr simple (bzw. unsichere) Form der Anmeldung
#my $token = $q->param('token');
#if ( $token eq '1795348177' ) {	# [en]	client is authenticated
#				# [de]	Client is berechtigt
#	$token = "token=$token";
#} else {			# [en]	client is not authenticated
#				# [en]	Client ist nicht berechtigt
#	print $q->header( -status => 404 );
#	exit 0;
#}

printf <<EOS, Sys::Hostname::hostname();
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/x
html1/DTD/xhtml1-strict.dtd">
<html xmlns='http://www.w3.org/1999/xhtml'>
 <head>
  <title>myCDN | %s</title>
  <meta name="format-detection" content="telephone=no">
EOS

if ( $section eq 'graph' ) {
	print <<EOS;
  <link rel="stylesheet" type="text/css" href="/jarmon/assets/css/jarmon_cdn.css" />
  <link rel="stylesheet" type="text/css" href="/jarmon/assets/css/tabs-no-images.css" />
  <link rel="stylesheet" type="text/css" href="/jarmon/assets/css/jquerytools.dateinput.skin1.css" />
EOS
}

print <<EOS;
  <style type="text/css">
	[data-role="page"] {
		font-family: "Verdana", "Arial", "Helvetica", sans-serif;
		font-weight: normal;
		left: 0;
		right: 0;
		top: 0;
		width: 100%;
		position:absolute; 
		background-color: white;
	}

	[data-role="header"] {
		height: 30px;
		color: $color_light;
		background-color: $color_dark;
		padding: 32px 0 0 2em;
		border-bottom: 8px solid $color_two;
	}
	[data-role="header"] h1 {
		margin-top: -.4em;
		color: $color_light;
		font-size: 200%;
		letter-spacing: .2em
	}

	[data-role="content"] {
		padding: 0;
		font-family: "Verdana", "Arial", "Helvetica", sans-serif;
		font-weight: normal;
		font-size: 18px;
		margin: 10px 4% 10px 4%;
		width: 92%;
		background-color: white;
		min-height: 410px;
		color: black;
	}

	[data-role="navbar"] ul {
		margin: 0;
		padding: 0;
		height: 22px;
	}
	[data-role="navbar"] ul li {
		font-size: 16px;
		list-style: none;
		display: inline;
		margin: .2em;
		padding: 0;
	}
	[data-role="navbar"] {
		margin: -2.2em 1.5em 20px 0;
		float: right;
		vertical-align: top;
	}
	[data-role="navbar"] a:link,
	[data-role="navbar"] a:visited,
	[data-role="navbar"] a:hover,
	[data-role="navbar"] a:active,
	.heading {
		color: $color_light;
		padding: 0 .5em;
		text-decoration: none;
	}
	[data-role="navbar"] a.ui-btn-active {
		color: white;
		border: 8px solid $color_two;
		padding-bottom: 8px;
		background-color: $color_two;
		border-radius: 8px 8px 0 0;
	}

	h1, h2, h3, h4, legend {
		font-family: 'Lucida Grande', Geneva, Arial, Verdana, sans-serif; 
		font-weight: 400;
		margin-top: 0.9em;
		margin-bottom: 0;
		padding-bottom: 0;
		color: $color_dark;
		letter-spacing: 0.04em;
	}

	[data-role="content"] h2 {
		border-bottom: 1px solid $color_two;
		margin-bottom: .4em;
	}

	[data-role="content"] a:link,
	[data-role="content"] a:visited,
	[data-role="content"] a:hover,
	[data-role="content"] a:active {
		color: $color_two;
		text-decoration: none;
	}

	p.code {
		font-size: 70%;
		font-family: monospace;
		white-space: pre;
		padding-bottom: 0;
	}

	[data-role="footer"] {
		padding: 0 0 0 10em;
	}
	[data-role="footer"] h4 {
		padding: .8em 3em .8em 0;
		border-top: 9px solid $color_two;
		border-radius: 9px 0 0 28px;
		background-color: $color_one;
		color: $color_light;
		font-size: 12px;
		text-align: right;
	}

	p.error, p.warning, p.ok {
		color: white;
		border-radius: 8px;
		padding: 8px 10px 8px 10px;
		margin: .5em 1em .8em 0.2em;
		width: 92%;
		font-size: 20px;
		text-decoration: none;
		display: block;
	}
	p.error {
		border: 1px solid red;
		background-color: #FF8686;
	}
	p.ok {
		border: 1px solid green;
		background-color: #9CD257;
	}
	p.warning {
		border: 1px solid yellow;
		background-color: #FDC83A;
	}

	table.interface {
		border-collapse: collapse;
		font-size: 90%;
	}
	table.interface td {
		padding: .3em .8em;
		border: 1px solid $color_one;
	}
  </style>
EOS

# jarmon CSS code
if ( $section eq 'graph' ) {
	print <<EOS;
  <script type="text/javascript" src="/jarmon/assets/js/dependencies.js"></script>
  <script type="text/javascript" src="/jarmon/jarmon.js"></script>
  <script type="text/javascript" src="/jarmon/jarmon_cdn.js"></script>
  <script type="text/javascript">
    \$(function() {
      jarmon.buildTabbedChartUi(
        \$('.chart-container').remove(),
        jarmon.CHART_RECIPES_COLLECTD,
        \$('.tabbed-chart-interface'),
        jarmon.TAB_RECIPES_STANDARD,
        \$('.chartRangeControl')
      );
    });
  </script>
EOS
}

print	" </head>\n\n".
	" <body>\n".
	"  <div data-role='page'>\n".
	"   <div data-role='header' data-theme='a'>\n".
	"    <h1><a class='heading' border='0' href='$ENV{SCRIPT_NAME}?$token'><i>my</i>CDN</a></h1>\n".
	"   </div>\n";

# sections
print	"   <div data-role='navbar'>\n".
	"    <ul>\n";
for my $link ( qw/home XtreemFS PowerDNS lsyncd Network graph log help/ ) {
	# [en]	hide link to jarmon if no jarmon environment was found
	# [de]	Link zu jarmon verstecken, wenn jarmon nicht vorhanden ist
	if ( $link eq 'graph' and ! 
	     -e $ENV{'DOCUMENT_ROOT'} . 'jarmon/jarmon.js' ) {
		next;
	}

	my $class = '';

	if ( $ENV{'QUERY_STRING'} =~ /section=(\w+)/ ) {
		if ( $1 eq lc $link ) {
			$class = 'class="ui-btn-active" ';
		}
	} elsif ( $link eq 'home' and
		  ( !$ENV{'QUERY_STRING'} or $ENV{'QUERY_STRING'} =~ /^token=\d+$/ ) ) {
		$class = 'class="ui-btn-active" ';
	}

	printf	"     <li><a %s href='?section=%s&%s'>%s</a>\n",
		$class, lc $link, $token, $link eq 'help' ? '?' : $link;
}
print	"    </ul>\n".
	"   </div>\n";


# [en]	begin main content
# [de]	hier kommt die wesentliche Inhalt
print	"   <div data-role='content'>\n";


# [en]	web page with generic information
# [de]	Webseite mit allgemeinen Informationen
if ( !$section or $section eq 'home' ) {
	print	"  <h2>System information</h2>\n";
	if ( my $r_proc_cpuinfo = get_file_content( '/proc/cpuinfo', 'processor' ) ) {
		my $number_of_cores = scalar @$r_proc_cpuinfo;
		printf	"  This system has %d CPU core%s",
			$number_of_cores, $number_of_cores > 1 ? 's' : '';
	}
	if ( my $r_proc_meminfo = get_file_content( '/proc/meminfo', 'MemTotal' )) {
		if ( $$r_proc_meminfo[0] =~ /^MemTotal:\s+(\d+)\s/ ) {
			my $memory_kb = $1;
			if ( $memory_kb > 1000000 ) {	# [en] more than 1 GB
							# [de] mehr als 1 GB
				printf	" and %d GB of system memory.",
					$memory_kb / 1000 / 1000;
			} else {	# [en] memory is below 1 GB
					# [de] Arbeitsspeicher unter 1 GB
				printf	" and %d MB of system memory.",
					$memory_kb / 1000;
			}
		} else {
			print	".";
		}
	}
	print	"  <br/>\n";

	# [en]	provide some information from /proc
	# [de]	Zeige ein paar Werte aus dem /proc System
	my $linux_version = ${ get_file_content( '/proc/version' ) }[0];
	$linux_version =~ s/\(.*//;		# [en] remove some dust
						# [de] entferne den Müll
	print	"  $linux_version<br/>\n";

	printf	"  Average load is %s<br/>\n",
		${ get_file_content( '/proc/loadavg' ) }[0];

	my $uptime = (split( /\s+/, ${ get_file_content( '/proc/uptime' ) }[0] ))[0];
	printf	"  System is running since %s (%d days, %0.1f hours)<br/><br/>\n",
		timef( time() - $uptime ),
		int( $uptime/60/60/24 ), 
		$uptime/60/60 - int( $uptime/60/60/24 )*24;

	# server information
	printf(	"  Server <code>%s</code> is located in %s, %s ".
	        "<img src='%s/%s.svg' width='24' alt='country flag icon' /></h1><br/>\n",
			Sys::Hostname::hostname(),
			$ENV{'CDN_COUNTRY_NAME'}, $ENV{'CDN_COUNTRY_CODE'},
			$url_country_flag, lc $ENV{'CDN_COUNTRY_CODE'}
	);

	my $gi = Geo::IP->open( "/usr/share/GeoIP/GeoLiteCity.dat", GEOIP_STANDARD );
	my $record = $gi->record_by_addr( $ENV{'REMOTE_ADDR'} );
	#my $record = $gi->record_by_addr( '59.123.123.1' );

	# client information
	printf "  Client <code>%s</code> ", $ENV{'REMOTE_ADDR'};
	if ( $record ) {
		printf( "is accessing from %s, %s ".
			"<img src='%s/%s.svg' width='24' alt='flag icon' /><br/>\n",
				$record->country_name, $record->country_code,
				$url_country_flag, lc $record->country_code
		);

		print	"<p style='color:gray; font-size:80%'>\n";
		printf( "country code:         %s<br/>\n", $record->country_code );
		printf( "country code 3:       %s<br/>\n", $record->country_code3 );
		printf( "country name:         %s<br/>\n", $record->country_name );
		printf( "country regions:      %s<br/>\n", $record->region );
		printf( "country regions name: %s<br/>\n", $record->region_name );
		printf( "country city:         %s<br/>\n", $record->city );
		printf( "country postal code:  %s<br/>\n", $record->postal_code );
		printf( "country latitude:     %s<br/>\n", $record->latitude );
		printf( "country longitude:    %s<br/>\n", $record->longitude );
		printf( "country time zone:    %s<br/>\n", $record->time_zone );
		printf( "country area code:    %s<br/>\n", $record->area_code );
		printf( "country continent:    %s<br/>\n", $record->continent_code );
		printf( "country metro code:   %s<br/>\n", $record->metro_code );
		print	"</p>\n";

	} else {
		print " (no geo information available)<br/>\n";
	}


###  xtreemfs  ###
} elsif ( $section eq 'xtreemfs' ) {

	print	"  <h2>XtreemFS</h2>\n";

	# version info
	my $xtfs_version = `$path_to_xtfsutil --version`;
	$xtfs_version =~ s/xtfsutil\s*//;
	$xtfs_version =~ s!\((.*)\)!<span style='color:gray'>($1)</span>!;
	print 'version ' . $xtfs_version;

	# [en]	query XtreemFS utils and show output
	# [de]	zeige Informationen über das XtreemFS Tool
	print   "  <h3>local volumes</h3>\n";
	my @xtfs = `sudo $path_to_lsfs cdn-de-1`;
	print_cli_output( $?, $path_to_lsfs, \@xtfs );

	if ( my $r_mounts = get_file_content( '/proc/mounts', 'xtreemfs@' ) ) {
		for my $mount ( @$r_mounts ) {
			my ( $volume, $mountpoint, $fs, $options ) = split( /\s+/, $mount );
			$volume =~ s/^xtreemfs@//;	# remove prefix

			print   "  <h3>volume <b>$volume</b></h3>\n";
			my @xtfs = `sudo $path_to_xtfsutil $mountpoint`;
			print_cli_output( $?, $path_to_xtfsutil, \@xtfs );
		}
	}


###  PowerDNS  ###
} elsif ( $section eq 'powerdns' ) {

	print	"    <h2>PowerDNS</h2>\n";

	# [en] check if daemon is running
	# [de] Prüfe ob der Dienst läuft
	my $pdns_status = `sudo $path_to_pdns_control status`;
	if ( $pdns_status =~ /Child running on pid/ ) {

		# version info
		my $pdns_version = `sudo $path_to_pdns_control version`;
		printf "version %s<br/>\n", $pdns_version;

		my $pdns_stats = `sudo $path_to_pdns_control list`;
		# http://linux.die.net/man/8/pdns_control

		chomp( $pdns_stats );
		for my $attr ( split( ',', $pdns_stats ) ) {
			printf	"    %s: %s<br/>\n", split( '=', $attr );
		}
	} else {
		print   "  <p class='error'>Process not running</p>\n";
	}


###  lsyncd  ###
} elsif ( $section eq 'lsyncd' ) {

	print	"    <h2>lsyncd</h2>\n";
	if ( my $pid = get_pid_of('lsyncd') ) {
		if ( my $lsyncd_status = get_file_content( $lsyncd_status_file ) ) {
			printf  "  <p>\n";
			for my $line ( @$lsyncd_status ) {
				chomp( $line );
				print "$line<br/>\n";
			}
			printf  "  </p>\n";
		} else {
			print   "  <p class='error'>Error reading <code>".
				"$lsyncd_status_file</code></p>\n";
		}
	} else {
		print   "  <p class='error'>Process not running</p>\n";
	}


###  Network   ###
} elsif ( $section eq 'network' ) {

	print	"  <h2>Interfaces</h2>\n";

	if ( my $r_proc_dev = get_file_content( '/proc/net/dev' ) ) {

		# [en] show formatted config file
		# [de] zeige die schick formatierte Konfigurationdatei
		print   "  <table class='interface'>\n".
			"   <thead>\n".
			"    <tr>\n".
			"     <th>Status</th>\n".
			"     <th>Interface</th>\n".
			"     <th>Receive</th>\n".
			"     <th>Transmit</th>\n".
			"    </tr>\n".
			"   </thead>\n".
			"   <tbody>\n";

                foreach ( sort @$r_proc_dev ) {

			# [en] skip heading
			# [de] Überschriften dürfen übersprungen werden
			next if substr( $_,0,7 ) eq 'Inter-|';
			next if substr( $_,0,7 ) eq ' face |';
			s/^\s+//;	# [en] remove leading whitespace
					# [de] führende Leerzeichen löschen
			next if /^(br0|lo|tun)/;# [en] skip loopback and bridge
						# [de] Loopback und Bridge ignorieren

			# [en] get values and counters
			# [en] Werte und Zählerstände auslesen
			my ( $iface, @rx, @tx );
			( $iface, @rx[ 0..7 ], @tx[ 0..7 ] ) = split( /[:\s]+/, $_ );
			$iface =~ s/:$//;	# [en] remove trailing colon
						# [de] Doppelpunkt am Ende löschen
			$iface =~ s/:\d+$//;	# [en] remove strange numbers
						# [de] unbekannte Zahlen entfernen

			# [en] numbering (as defined by /proc/net/dev)
			# [de] Nummerierung (passend zu /proc/net/dev)
			#  0 - bytes
			#  1 - packets
			#  2 - errors
			#  3 - drops
			#  4 - fifo
			#  5 - frame
			#  6 - compressed
			#  7 - multicast

			# [en] get information about interface state from mii-tool
			# [de] Informationen über Netzwerkkarte vom mii-tool
			my $r_mii_tool = `sudo $path_to_miitool $iface`;
			my ( $state, $speed, $duplex ) = ( '', 'n/a', 'n/a' );
			if ( $r_mii_tool =~ /-(\w)D,/ ) {	# duplex
				$duplex = $1 eq 'F' ? 'full' : 'half';
			}
			if ( $r_mii_tool =~ /(1\d+)base/ ) {	# speed
				$speed = $1;
			}
			if ( $r_mii_tool =~ /link ok/ ) {	# state
				$state = 'up';
			} elsif ( $r_mii_tool =~ /no link/ ) {
				$state = 'down';
			}

			# MAC address
			my $mac = @{ get_file_content(
				"/sys/class/net/$iface/address") }[ 0 ];

                        printf( "    <tr>\n".
                                "     <td>%s<br/>[%s]</td>\n".
                                "     <td>\n".
				"      MAC: %s<br/>\n".
				"      Duplex: %s<br/>\n".
				"      Speed: %s\n".
				"     </td>\n".
                                "     <td>\n".
				"      MBytes: %.1f<br/>\n".
				"      Errors: %d<br/>\n".
				"      Drops: %d\n".
				"     </td>\n".
                                "     <td>\n".
				"      MBytes: %.1f<br/>\n".
				"      Errors: %d<br/>\n".
				"      Drops: %d\n".
				"     </td>\n".
                                "    </tr>\n",

				$state, $iface,
				$mac || 'unknown', $duplex , $speed,
				$rx[0] /1024 /1024, $rx[2], $rx[3], 
				$tx[0] /1024 /1024, $tx[2], $tx[3], 
			);
                }
		print	"   </tbody>\n".
			"  </table>\n\n";
	}


	print	"  <h2>NetFlow</h2>\n";
	if ( my $pid = get_pid_of('softflowd') ) {
		if ( my $softflowd_conf = get_file_content( $conf_softflowd ) ) {
			printf	"  <p>\n";
			for my $line ( @$softflowd_conf ) {
				if ( $line =~ /-n (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ ) {
					printf	"   Exporter IP: $1<br/>\n";
				}
				if ( $line =~ /-v (\d)/ ) {
					printf	"   NetFlow version: $1<br/>\n";
				}
				if ( $line =~ /INTERFACE\s*=s*"([\w\d]+)"/ ) {
					print	"   Inspecting interface: $1<br/>\n";
				}
			}
			printf	"  </p>\n";
		} else {
			print   "  <p class='error'>Error reading <code>".
				"$conf_softflowd</code></p>\n";
		}
	} else {
		print   "  <p class='error'>Process not running</p>\n";
	}
	

###  Manuals   ###
} elsif ( $section eq 'help' ) {

	print	"  <h2>About</h2>\n".
		"  <p>\n".
		"   my &mdash; <i>Content Delivery Network</i><br/>\n".
		"   &copy; 2015, 2016\n".
		"  </p>\n";

	print	"  <h2>Additional information</h2>\n".
		"  <p>\n".
		"   <a href='http://$ENV{HTTP_HOST}:30636/'>XtreemFS MRC</a> &ndash; Metadata and Replica Catalog<br/>\n".
		"   <a href='http://$ENV{HTTP_HOST}:30638/'>XtreemFS DIR</a> &ndash; Directory Service<br/>\n".
		"   <a href='http://$ENV{HTTP_HOST}:30640/'>XtreemFS OSD</a> &ndash; Object Storage Device<br/>\n".
		"   <a href='http://www.xtreemfs.org/xtfs-guide-1.5.1/index.html'>XtreemFS User Guide</a><br/>\n".
		"   <a href='https://doc.powerdns.com/'>PowerDNS documentation</a><br/>\n".
		"   <a href='https://github.com/axkibe/lsyncd'>Live Syncing (Mirror) Daemon</a><br/>\n".
		"  </p>\n";




###  graph     ###
} elsif ( $section eq 'graph' ) {
	print <<EOS;
        <div class="chartRangeControl">
            <form>
                <div>
                    <span class="timerange_control custom">
                        <img src="/jarmon/assets/icons/calendar.png" width="16" height="16" alt="calendar" class="from_custom"
                             title="Click to choose a custom start date" />
                        <input name="from_custom" type="text" readonly="readonly"
                               title="Time range start" />
                        <img src="/jarmon/assets/icons/calendar.png" width="16" height="16" alt="calendar" class="to_custom"
                             title="Click to choose a custom end date" />
                        <input name="to_custom" type="text" readonly="readonly"
                               title="Time range end" />
                    </span>
                    <span class="timerange_control standard">
                        <select name="from_standard"
                                title="Time range shortcuts - click to select an alternative time range" >
                        </select>
                    </span>
                    <input name="from" type="hidden"  />
                    <input name="to" type="hidden"  />
                    <select name="tzoffset"
                            title="Timezone offset - click to choose a custom timezone offset" ></select>
                    <input name="action" value="Update" type="button"
                           title="Graph update - click to update all graphs" />
                </div>
                <div class="range-preview"
                     title="Time range preview - click and drag to select a custom timerange" ></div>
            </form>
        </div>

        <div class="tabbed-chart-interface"></div>
        <div class="chart-container">
            <div class="chart-header">
                <h2 class="title"></h2>
            </div>
            <div class="error"></div>
            <div class="chart"></div>
            <div class="graph-legend"></div>
        </div>
EOS


###  logging   ###
} elsif ( $section eq 'log' ) {
	print	"  <h2>Logging</h2>\n";
	my $file = $q->param('file');

	my %query_type = (
		'0001' => 'A',
		'0002' => 'NS',
		'0005' => 'CNAME',
		'0006' => 'SOA',
		'000c' => 'PTR',
		'000f' => 'MX',
		'0010' => 'TXT',
		'001c' => 'AAAA',
		'0026' => 'A6',
		'00fb' => 'IXFR',
		'00fc' => 'AXFR',
		'00ff' => 'wildcard',
	);

	if ( $file ) {
		if ( -e "/var/log/$file" ) {
			if ( open( my $fh_log, '<', "/var/log/$file" )) {
				my $line_counter = 0;
				LINE:
				while ( my $line = <$fh_log> ) {
					chomp( $line );

					# special handling for geoipdns log
					# http://www.dqd.com/~mayoff/notes/djbdns/tinydns-log.html
					if ( $file eq 'geoipdns.log'
					     and $line =~ /^[0-9a-f]{8}:[0-9a-f]{4}:[0-9a-f]{4}/ ) {
						print	"  <span style='color:gray'>\n";

						# 7f000001:b323:06e7 + 0001 www-cdn.stubbig.com
						my ( $socket, $sep, $type, $name ) = split( /\s+/, $line );
						my ( $ip, $port, $id ) = split( /:/, $socket );

						# convert hex to decimal
						$ip = sprintf "%d.%d.%d.%d",
							hex( substr( $ip,0,2 ) ), hex( substr( $ip,2,2 ) ),
							hex( substr( $ip,4,2 ) ), hex( substr( $ip,6,2 ) );
						$port = hex( $port );

						printf( "   query type %s for <code>%s</code><br/>\n",
							$query_type{ $type }, $name );
						print "   client socket $ip:$port<br/>\n";
						print "   request was ";
						if ( $sep eq '+' ) {
							print "answered properly";
							# make the log line green (RdYlGn-7-div-7)
							$line = "<span style='color:#1a9850'>$line</span>";
						} elsif ( $sep eq '-' ) {
							print "dropped (no authority to answer)";
							# make the log line red (RdYlGn-7-div-1)
							$line = "<span style='color:#d73027'>$line</span>";
						} elsif ( $sep eq 'I' ) {
							print "dropped (type is not implemented)";
						} elsif ( $sep eq 'C' ) {
							print "dropped (invalid/unknown class)";
						} elsif ( $sep eq '/' ) {
							print "dropped (0 or >1 query)";
						} else {
							print "dropped (unknown error)";
						}
						print "\n  </span><br/>\n";

					} elsif ( substr( $file,0,8 ) eq 'xtreemfs' ) {
						if ( substr( $line,0,5 ) eq '[ E |' ) {
							# [en] make the log line red (RdYlGn-7-div-1)
							# [de] diese Zeile wird rot  (RdYlGn-7-div-1)
							$line = "<span style='color:#d73027'>$line</span>";
						} elsif ( substr( $line,0,5 ) eq '[ W |' ) {
							# [en] make the log line orange (RdYlGn-11-div-4)
							# [de] diese Zeile wird orange  (RdYlGn-11-div-4)
							$line = "<span style='color:#fdae61'>$line</span>";
						} elsif ( substr( $line,0,1 ) eq '#' ) {
							$line = "<span style='color:gray'>$line</span>";
						}
					}

					print "  $line<br/>\n";

					# [en] stop processing too long log files
					# [de] Abbruch nach dem festgesetzten Limit
					if ( $line_counter > $logfile_max_lines ) {
						printf( "  <br/>\n".
						  "  <i>log file exceeds %d entries, ".
						  "skipping remaining lines...</i>\n",
							$logfile_max_lines );
						last LINE;
					}
					$line_counter++;
				}
				close $fh_log;

				if ( $line_counter == 0 ) {
					print "  <i>empty</i>\n";
				}
			} else {
				print   "  <p class='error'>Error opening ".
					"<code>/var/log/$file</code>: $!</p>\n";
			}
		} else {
			print   "  <p class='error'>File <code>/var/log/$file</code>".
				" not found</p>\n";
		}

	} else {
		print   "  <table style='border-spacing: .9em 0em;'>\n".
			"   <thead>\n".
			"    <tr>\n".
			"     <th>filename</th>\n".
			"     <th>size</th>\n".
			"     <th>timestamp</th>\n".
			"    </tr>\n".
			"   </thead>\n".
			"   <tbody>\n";

		for my $file ( qw!messages lsyncd.log xtreemfs/dir.log xtreemfs/mrc.log xtreemfs/osd.log! ) {
			# [en] last file access
			# [de] letzter Zugriff auf diese Datei
			my $file_last_access = (stat("/var/log/$file"))[ 9 ];
			next if time() - $file_last_access > 7 * 24 * 60 * 60;

			# [en] get file size
			# [de] Dateigröße ermitteln
			my $file_size = (stat( "/var/log/$file" ))[ 7 ];
			if ( $file_size > 1000*1000 ) {
				$file_size = sprintf "%0d KB", $file_size / 1024;
			} elsif ( $file_size > 0 ) {
				$file_size = sprintf "%0.1f KB", $file_size / 1024;
			}

			printf( "     <tr>\n".
				"      <td><a href='%s&file=%s'>%s</a></td>\n".
				"      <td>%s</td>\n".
				"      <td>%s</td>\n".
				"     </tr>\n",
					$ENV{'REQUEST_URI'}, $file, $file, 
					$file_size, timef( $file_last_access )
			);
		}
		print	"   </tbody>\n".
                	"  </table>\n";

	}


###  unknown   ###
} else {
	print	"  <br/><p class='error'>unknown section '$section'</p>\n";
}

for ( sort keys %ENV ) {
	next if $_ eq 'SERVER_SIGNATURE';
	next if /_proxy$/;		# [en] don't expose proxy credentials here
					# [de] Proxy-Anmeldeinformationen verstecken
	printf	"  <!-- %s: %s -->\n", $_, $ENV{ $_ };
}


# [en] end main content
# [de] hier endet das Hauptdokument
print	"   </div>\n";


# [en] footer with timestamp
# [de] Fußzeile mit Datum und Uhrzeit
print	"   <div data-role='footer'>\n";
$_ = $ENV{'SCRIPT_FILENAME'};
s!.*/!!;	# [en] remove path
		# [de] Pfad entfernen
printf	"    <h4>%s &ndash; %s &ndash; %s</h4>\n",
	$_, Sys::Hostname::hostname(), timef( time() );
print	"   </div>\n";

quitProgram( 0 );



=head2 Methods

=over 12

=item I<quitProgram> I<(Exitcode)>

[en] Finalize the HTML page and quit with an exit code.
Returns nothing.

[de] Beendet die HTML-Seite und liefert den Exitcode.
Kein Rueckgabewert.

=cut
sub quitProgram( $ ) {
	print	" </body>\n</html>";
	exit( 0 );
}


=item I<print_cli_output> I<(Errorcode, Kommando, Kommandoausgabe)>

[en] Print the output of a CLI command in nice HTML format.
This method expects an error code, the CLI command and its output
as reference.
Returns nothing.

[de] Formartiert die Ausgabe von einem Kommando passend fuer eine
HTML-Webseite. Erwartet einen Errorcode, das ausgefuehrte Kommando
und seine Ausgabe (als Referenz).
Kein Rueckgabewert.

=cut
sub print_cli_output( $$$ ) {
	my ( $err_no, $command, $r_output ) = @_;

	if ( $err_no ) {
		printf	"  <p class='error'>Error %s while talking ".
			"to <code>%s</code></p>\n",
			$err_no, $command ;

		print	"make sure you have modified <code>sudo</code> ".
			"with something like<br/>\n".
			"<code>apache  ALL=(ALL)  NOPASSWD:$command</code>\n";

	} else {
		print	"  <p class='code'>";

		for ( @$r_output ) {
			chomp;
			next if /^\s*$/;	# skip blank lines
			next if /^-{66}$/;	# skip long dash

			print "$_\n";
		}
		print	"  </p>\n\n";
	}
	return;
}


=item I<timef> I<(Unix-Zeitangabe)>

[en] Convert a UNIX timestamp (seconds since 1.1.1970) to ISO 8601

Returns is the human readable timestamp.

[de] Formartiert einen UNIX-Zeitstempel (Sekunden seit 1.1.1970)
nach ISO 8601.

Rueckgabewert ist der formatierte Zeitstempel.

=cut
sub timef( $ ) {
	my ($sec,$min,$hour,$day,$mon,$year,$yday,$tz) = localtime( $_[0] );
	return sprintf( "%04d-%02d-%02d %02d:%02d:%02d",
		$year+1900, $mon+1, $day, $hour, $min, $sec );
}


=item I<get_pid_of> I<(Prozessname)>

[en] Find the process name in the list of processes and extract
the process ID.

Returns the process ID or nothing if process wasn't found.

[de] Sucht in der Prozessliste nach dem angegebenen Prozessnamen und
liefert die Prozessnummer zuueck.

Rueckgabewert ist die Prozessnummer oder nichts wenn kein Prozess
gefunden wurde.

=cut
sub get_pid_of( $ ) {
	my ( $process_name ) = @_;
	
	# [en] shell command to get process id
	# [de] Kommando zum Ermitteln der Process ID
	my $cmd = sprintf( "ps -e|grep [%s]%s|awk '{print \$1}'",
		substr( $process_name, 0, 1 ),	# first letter
		substr( $process_name, 1, length( $process_name ) )
	);

	my $pid = `$cmd`;
	chomp( $pid );
	return $pid;
}


=item I<get_file_content> I<(Dateiname, Regulaerer Ausdruck)>

[en] Open a file and return its content as reference on an array.
An optional regular expression is used to filter the output.

Returns a reference to array or undef on error.

[de] Oeffnet eine Datei und liefert den Inhalt als Referenz auf ein
Array zurueck. Bei Angabe eines optionalen regulaeren Ausdrucks
werden nur die Zeilen behandelt, die auf den Ausdruck passen.

Rueckgabewert ist eine Referenz oder 'undef' im Fehlerfall.

=cut
sub get_file_content( $@ ) {
	my ( $filename, $regex ) = @_;

	if ( open( my $fh, '<', $filename )) {
		my @content;
		while ( my $line = <$fh> ) {
			chomp( $line );
			if ( $regex ) {
				if ( $line =~ /$regex/ ) {
					push @content, $line;
				}
			} else {
				push @content, $line;
			}
		}
		close $fh;
		return \@content;
	} else {	# file open failed
		return undef;
	}
}


=item I<get_country_code> I<(IP Adresse)>

[en] Find the country where the given IP address is located.
This method expects an IPv4 address. Return the ISO 3166 country
code or undef (on error)

[de] Ermittle in welchem Land sich die IP-Adresse befindet.
Diese Methode erwartet eine IPv4-Adresse. Rueckgabewert ist ein
ISO 3166 Landeskennung oder undefiniert (bei Fehler).

=cut
sub get_country_code ($) {
	my ( $ip ) = @_;

	# [en] skip country determination if IP is from non-public block (RFC 1918)
	# [de] die nichtöffentlichen IP-Bereiche gehörten zu keinem Land (RFC 1918)
	if (	in_subnet( '10.0.0.0/8',     $ip )
	     or in_subnet( '172.16.0.0/12',  $ip )
	     or in_subnet( '192.168.0.0/16', $ip )
	     or in_subnet( '127.0.0.0/8',    $ip ) ) {
		return undef;

	} else {
		# [en] get country information from public IP address
		# [de] suche die Landeskennung für öffentliche IP-Adressen
		my $gi = Geo::IP->open(
			"/usr/share/GeoIP/GeoLiteCity.dat", 
			GEOIP_STANDARD );
		my $record = $gi->record_by_addr( $ip );

		return $record ? $record->country_code : undef;
	}
}


=item I<in_subnet> I<(subnet, ip address)>

[en] Checks to see if IP address is in subnet. Subnet is defined as single IP
address, or CIDR block. Partial CIDR format (192.168/16) is NOT supported

Diese Methode erwartet eine IPv4-Range und eine einzelne IPv4-Adresse.
Rueckgabewert ist Wahrheitswert.

HINT: This subroutine (including help text) was not written be me.
      It is adopted from netdisco version 0.94 (www.netdisco.org).

[de] Prüfe ob die angegebene IP-Adresse zum Subnetz passt. Das Subnetz ist
entweder eine IP-Adresse, oder ein CIDR-Block. Unvollständige CIDR-Angaben
(z.B. 192.168/16) werden nicht unterstützt.

This method expects an IPv4 range and a single IPv4 address. Returns boolean.

HINWEIS: Diese Subroutine gehört zur Software netdisco Version 0.94

=cut
sub in_subnet ($$) {
	my ( $subnet, $ip ) = @_;

	unless ( defined $subnet and defined $ip ) {
		return undef;
	}

	# No / just see if they're equal
	if ( $subnet !~ m!/! ) {
		return ( $subnet eq $ip ) ? 1 : 0;
	}

	# Parse Subnet
	my ($root,$bits);
	if ( $subnet =~ /^([\d\.]+)\/(\d+)$/ ) {
		$root = $1;  $bits = $2;
	} else {
		return undef;
	}

	my $root_bin = unpack( "N", pack( "C4", split( /\./, $root )));
	my $mask = 2**32 - (2 ** (32 - $bits));

	# Parse IP
	my $ip_bin = unpack( "N", pack( "C4", split( /\./, $ip )));

	return (($ip_bin & $mask) == ($root_bin & $mask)) ? 1 : 0;
}


=item I<get_country_url> I<(IP Adresse)>

[de] Ermittle in welchem Land sich die IP-Adresse befindet und formatieren das
Ergebnis as <img> Object.
Diese Methode erwartet eine IPv4-Adresse. Rueckgabewert ist ein HTML <img>
Object mit ISO 3166 Countrycode oder ein leerer String (bei Fehler).

=cut
sub get_country_url ($) {
	my ( $ip ) = @_;

	my $country_code = get_country_code( $ip );
	if ( $country_code ) {
		return sprintf( "<img src='%s/%s.svg' width='20' alt='country flag' />",
			$url_country_flag, lc $country_code
		);
	} else {
		# [en]	dont return undef here, since result will be used in scalar
		# [de]	hier muss ein Wert zurückgeliefert werden, weil das Ergebnis später
		# 	in einer Skalar-Variable gespeichert wird
		return '';
	}
}


__END__

=back

=head1 BUGS

[en] None known.
[de] Keine bekannt.

=head1 AUTOR

Markus Stubbig

=head1 SEE ALSO

L<pdns_control>, L<xtfsutil(1)>, L<softflowd(8)>, L<lsyncd(1)>

=cut
